#!/bin/sh
# \
exec wish -f "$0" ${1+"$@"}

# Amateur Radio Field Day Database Server
# Uses socket communications between server & local/remote clients

#=======================================================================
# support code for tcl/tk socket services
#=======================================================================

namespace eval ServerSocket {
	namespace export ServerSocket SSclose SSactivate SSwrite SSbroadcast SSdrop

	variable _socket
	variable _clients
	variable _serverFunction ""
}

proc ServerSocket::ServerSocket {{port_ 20002}} {
	variable _socket
	# Open the physical socket.
	if {[catch {socket -server ServerSocket::accept $port_} _socket]} {
		return -code error
	}
	return -code ok
}

proc ServerSocket::SSdrop {sd} {
variable _clients
	unset _clients($sd)
	catch {puts $sd "close"}
	catch {close $sd}
}

proc ServerSocket::SSclose {} {
	variable _socket
	variable _clients
	variable _serverFunction

	catch {close $_socket}
	foreach sock [array names _clients] {
		ServerSocket::SSdrop $sock
	}
	set _serverFunction ""
}

proc ServerSocket::SSactivate {command_} {
variable _serverFunction
	set _serverFunction $command_
}

proc ServerSocket::SSwrite {sd_ data_} {
	if {[catch {puts $sd_ $data_} err]} {
		error $err
	}
}

proc ServerSocket::SSbroadcast {data_} {
	variable _clients
	foreach sock [array names _clients] {
		if {[catch {puts $sock $data_} err]} {
			error $err
		}
	}
}

proc ServerSocket::accept {sd_ address_ port_} {
variable _clients
	set _clients($sd_) [list $address_ $port_]
	fconfigure $sd_ -buffering line
	fileevent $sd_ readable [list ServerSocket::handleClientInput $sd_]
}

proc ServerSocket::handleClientInput {sd_} {
	variable _serverFunction
	variable _clients

	catch {gets $sd_ data} rc

	if {$rc < 0} {
		ServerSocket::SSdrop $sd_
		return
	}
	if {[string compare $data "quit"] == 0} {
		ServerSocket::SSclose
	} else {
		if {[llength $_serverFunction]} {
				eval {$_serverFunction $sd_ $data}
		} else {
			ServerSocket::SSbroadcast $data
		}
	}
}

namespace eval ClientSocket {
	namespace export ClientSocket CSwrite CSregister CSclose
	variable _socket
	variable _clientFunction ""
}

proc ClientSocket::CSclose {} {
	variable _socket
	close $_socket
}

proc ClientSocket::CSregister {command_} {
	variable _clientFunction
	set _clientFunction $command_
}

proc ClientSocket::ClientSocket {{host_ localhost} {port_ 20002}} {
	variable _socket
	if {[catch {socket $host_ $port_} _socket]} {
		return -code error "Server DOWN"
	}
	fconfigure $_socket -buffering line
	fileevent $_socket readable ClientSocket::handleServerInput
	return $_socket
}

proc ClientSocket::CSwrite {data_} {
	variable _socket
	puts $_socket $data_
}

proc ClientSocket::handleServerInput {} {
	variable _socket
	variable _clientFunction
	if {[gets $_socket data] < 0} {
		return
	} else {
		if {[string compare $data "close"] == 0} {
			ClientSocket::CSclose
			$_clientFunction "CLOSE"
			return
		}
		if {[llength $_clientFunction]} {
			$_clientFunction $data
		} else {
			puts $data
		}
	}
}

#=======================================================================

namespace import ServerSocket::*

#=======================================================================
# if config.fd exists these values are read from the file during init
# for example
#
# callsign:W1HKJ
# class:2A
# multiplier:5
# section:AL
# autosave:60    value = seconds
# address:dave-dc7600
# port:20001
# datafile:fd2016.log
#
# User is presented with a configuration dialog allowing these value
# to be changed for a specific usage
#
# Data store files are simple comma separated record files
#
# Each record is a comma separated list
#
# BAND,MODE,CALL,SECTION,CLASS,OP_CALL
# where:
#
# BAND     band on which contact made
# MODE     one of CW, DIG, or PHONE
# CALL     contacts call sign
# SECTION  contacts section
# CLASS    contacts class
# OP_CALL  call of operator making contact (local)
#=======================================================================

variable staCallsign STACALL
variable staMult 5
variable staClass 1A
variable staSection AL
variable socket_address "127.0.0.1"
variable socket_port 20002
variable autosave 60 # autosave every 60 seconds
variable datafile "STACALL.fd"

global users
global contacts
global fd_data
global needsync
global log
variable cumScore 0

global lblContacts
global lblOperator

variable bands {160 80 40 20 17 15 12 10 6 2 220 440}
variable modes {CW DIG PHONE}
global val
set val(CW) 2
set val(DIG) 2
set val(PHONE) 1

proc FD_server {sock line} {
	set args [split $line " "]
	set command [lindex $args 0]
	switch $command {
		LOGON {FD_logon $sock [lindex $args 1] [lindex $args 2]}
		LOGOFF {FD_logoff $sock}
		DUPCHECK {FD_dupcheck $sock $line}
		ADD {FD_add $sock $line}
		VIEW {}
		default {tk_messageBox -type ok "Invalid command\n$sock\n$line" -icon error}
	}
}

proc logentry {str} {
	global log
	$log config -state normal
	$log insert end $str
	$log config -state disabled
	$log see end
}

proc FD_display {str} {
	set sec [clock seconds]
	set szTime [clock format $sec -format "%D %H:%M:%S" -gmt true]
	logentry "$szTime $str\n"
}

proc FD_logon {sock operator bandmode} {
	global users log staCallsign staClass staMult staSection cumScore
	global lblContact lblOperator oper
	if {[info exists users($bandmode)]== 1} {
		SSwrite $sock "LOGON_DENIED"
		SSdrop $sock
	} else {
		set users($bandmode) $operator
		set users($sock) $bandmode
		$lblContact($bandmode) config -bg thistle
	$lblOperator($bandmode) config -bg thistle
	set oper($bandmode) $operator
		SSwrite $sock "LOGON_OK $staCallsign $staClass $staMult $staSection"
		SSwrite $sock "SCORE $cumScore"
		FD_display "$operator $bandmode LOGGED ON"
		FD_announce
	}
}

proc FD_logoff {sock} {
	global users log lblContact lblOperator oper
	if {[info exists users($sock)] == 1} {
		set bmode $users($sock)
		$lblContact($bmode) config -bg white
	$lblOperator($bmode) config -bg white
	set oper($bmode) ""
		FD_display "$users($bmode) $bmode LOGGED OFF"
		unset users($bmode)
		unset users($sock)
		SSwrite $sock "LOGOFF_OK"
		SSdrop $sock
	} else {
		SSwrite $sock "LOGOFF_NOP"
	}
	FD_announce
}

proc FD_dupcheck {sock data} {
	global fd_data users
	set callsign [lindex $data 1]
	set query "$users($sock),$callsign,*"
	set ix [lsearch $fd_data $query]
	if {$ix >= 0} {
		SSwrite $sock "DUP"
	} else {
		SSwrite $sock "NODUP"
	}
}

proc FD_announce {} {
	global contacts bands modes oper
	set worked {}
	foreach band $bands {
		foreach mode $modes {
			lappend worked "$band $mode $contacts($band,$mode) $oper($band,$mode)"
		}
	}
	SSbroadcast "WORKED $worked"
}

proc FD_add {sock data} {
	global fd_data needsync users
	set callsign [lindex $data 1]
	set query "$users($sock),$callsign*"
	set ix [lsearch $fd_data $query]
	if {$ix >= 0} {
		SSwrite $sock "REJECT"
	} else {
		set bandmode "$users($sock)"
		set newrec    "$bandmode,"
		append newrec "[lindex $data 1],"
		append newrec "[lindex $data 2],"
		append newrec "[lindex $data 3],"
		append newrec "$users($bandmode)"
		lappend fd_data $newrec
		UpdateScores $users($sock)
		set needsync 1
		SSwrite $sock "ACCEPT"
		FD_display $newrec
		FD_announce
	}
}

proc FD_readfile {fname} {
	global fd_data needsync

	if [catch {open $fname r} fid] {
	FD_display "Creating new log $fname"
	if [catch {open $fname w} fid] {
		FD_display "Cannot create $fname: $fid"
	} else {
		close $fid
	}
	} else {
		FD_init
		set tmp [read -nonewline $fid]
		close $fid
		set fd_data [split $tmp "\n"]
		foreach rec $fd_data {
			if {$rec == ""} {
				continue
			} else {
				set lrec [split $rec ","]
				UpdateScores "[lindex $lrec 0],[lindex $lrec 1]"
			}
		}
		FD_announce
	}
	set needsync 0
}

proc FD_writefile {fname} {
	global fd_data needsync
	set fdata [join $fd_data "\n"]
	if [catch {open $fname w} fid] {
		FD_display "Cannot write $fname: $fid"
	} else {
		puts -nonewline $fid $fdata
		close $fid
		set needsync 0
	}
}

# UpdateScores expect data to be of the form band,mode; ie 80,CW
# contacts information is kept in arrays which are indexed by that pair
proc UpdateScores {data} {
	global contacts
	global cumScore val staMult
	set band [lindex [split $data ","] 1]
	incr contacts($data)
	incr cumScore [expr $val($band)*$staMult]
	SSbroadcast "SCORE $cumScore"
}

proc FD_read_config {} {
	global staCallsign staMult staClass staSection
	global socket_address socket_port
	global autosave datafile
	if [catch {open "config.fd" r} fid] {
		FD_display "Cannot open config.fd"
	} else {
		set tmp [read $fid]
		close $fid
		set config [split $tmp "\n"]
		foreach rec $config {
			set lrec [split $rec ":"]
			set name "[lindex $lrec 0]"
			switch $name {
				callsign   { set staCallsign "[lindex $lrec 1]" }
				class      { set staClass "[lindex $lrec 1]" }
				multiplier { set staMult [lindex $lrec 1] }
				section    { set staSection "[lindex $lrec 1]" }
				address    { set socket_addr "[lindex $lrec 1]" }
				port       { set socket_port [lindex $lrec 1] }
				autosave   { set autosave [lindex $lrec 1] }
				datafile   { set datafile [lindex $lrec 1] }
				default    {}
			}
		}
	}
}

proc FD_write_config {} {
	global staCallsign staMult staClass staSection
	global socket_address socket_port
	global autosave datafile
	if [catch {open "config.fd" w} fid] {
		FD_display "Cannot write to config.fd"
	} else {
	puts $fid "callsign:$staCallsign"
	puts $fid "class:$staClass"
	puts $fid "multiplier:$staMult"
	puts $fid "section:$staSection"
	puts $fid "address:$socket_address"
	puts $fid "port:$socket_port"
	puts $fid "autosave:$autosave"
	puts $fid "datafile:$datafile"
		close $fid
	}
}

proc FD_init {} {
	global contacts fd_data bands modes needsync
	set needsync 0
	set fd_data {}
	foreach band $bands {
		foreach mode $modes {
			set contacts($band,$mode) 0
		}
	}
}

proc FD_savelog {} {
	global log
	set txt [string range [$log get 1.0 end] 0 end-1]
	set fid [open "fdlog.txt" w]
	puts -nonewline $fid $txt
	close $fid
}

proc FD_readlog {} {
	if [catch {open "fdlog.txt" r} fid] {
	} else {
		set txt [read $fid]
		close $fid
		logentry $txt
	}
}

proc FD_save_data {} {
	global datafile
	FD_writefile $datafile
	FD_savelog
}

proc FD_autosave {} {
	global autosave
	FD_save_data
	FD_display "Saving data file"
	after [expr 1000*$autosave] FD_autosave
}

proc FD_Quit {} {
global needsync datafile
	SSbroadcast "QUIT"
	SSclose
	if {$needsync == 1} {
		FD_writefile $datafile
	}
	FD_savelog
	exit
}

proc FD_exit {} {
	set ans [tk_messageBox -type okcancel -default cancel \
			-message "Closing Server will\nclose all clients" \
		-title "CLOSE DOWN SERVER" \
		-icon question]
	if {$ans == "ok"} {
		FD_Quit
	}
}

proc FD_About {} {
	tk_messageBox -type ok \
		-message "Created by W1HKJ" \
		-icon info
}

proc Scrolled_Text { f args } {
	frame $f
	eval {text $f.text -wrap none \
		-xscrollcommand [list $f.xscroll set] \
		-yscrollcommand [list $f.yscroll set]} $args
	scrollbar $f.xscroll -orient horizontal \
		-command [list $f.text xview]
	scrollbar $f.yscroll -orient vertical \
		-command [list $f.text yview]
	grid $f.text $f.yscroll -sticky news
	grid $f.xscroll -sticky news
	grid rowconfigure $f 0 -weight 1
	grid columnconfigure $f 0 -weight 1
	return $f.text
}

#=============================================================
# Procedures to help build dialogs.
#=============================================================

proc Dialog_Create {top title args} {
	global dialog
	if [winfo exists $top] {
		switch -- [wm state $top] {
			normal {
				# Raise a buried window
				raise $top
			}
			withdrawn -
			iconic {
				# Open and restore geometry
				wm deiconify $top
				catch {wm geometry $top $dialog(geo,$top)}
			}
		}
		return 0
	} else {
		eval {toplevel $top} $args
		wm title $top $title
		return 1
	}
}

proc Dialog_Wait {top varName {focus {}}} {
	upvar $varName var

	# Poke the variable if the user nukes the window
	bind $top <Destroy> [list set $varName cancel]

	# Grab focus for the dialog
	if {[string length $focus] == 0} {
		set focus $top
	}
	set old [focus -displayof $top]
	focus $focus
	catch {tkwait visibility $top}
	catch {grab $top}

	# Wait for the dialog to complete
	tkwait variable $varName
	catch {grab release $top}
	focus $old
}

proc Dialog_Dismiss {top} {
	global dialog
	# Save current size and position
	catch {
		# window may have been deleted
		set dialog(geo,$top) [wm geometry $top]
		wm withdraw $top
	}
}

#=============================================================
proc Scroll_Set {scrollbar geoCmd offset size} {

  if {$offset != 0.0 || $size != 1.0} {
    eval $geoCmd                     ;# Make sure it is visible
    $scrollbar set $offset $size
  } else {
    set manager [lindex $geoCmd 0]
    $manager forget $scrollbar       ;# hide it
  }
}

proc Scrolled_Listbox { f args } {

  frame $f
  listbox $f.list \
    -xscrollcommand [list Scroll_Set $f.xscroll \
      [list grid $f.xscroll -row 1 -column 0 -sticky we]] \
    -yscrollcommand [list Scroll_Set $f.yscroll \
      [list grid $f.yscroll -row 0 -column 1 -sticky ns]]

  eval {$f.list configure} $args
  scrollbar $f.xscroll -orient horizontal \
    -command [list $f.list xview]
  scrollbar $f.yscroll -orient vertical \
    -command [list $f.list yview]

  grid $f.list $f.yscroll -sticky news
  grid $f.xscroll -sticky news
  grid rowconfigure $f 0 -weight 1
  grid columnconfigure $f 0 -weight 1
  return $f.list
}

proc FD_View {which} {
	global fd_data
	set worked [lsort $fd_data]
	set byband {}
	if {$which == "All"} {
		set byband $worked
	} else {
		foreach rec $worked {
			if {$which == [lindex [split $rec ","] 0]} {
				lappend byband $rec
			}
		}
	}

	set f .vb
	toplevel $f
	wm withdraw $f
	wm title $f "View Contacts"

	set lstbox [Scrolled_Listbox $f.lst \
		-exportselection 0 \
		-background seashell1 \
		-selectmode single -setgrid 1 \
		-font Courier -width 50 -height 20]

	set line [format "%4s |%4s |%10s |%5s |%5s|%10s" \
			"Band" "Mode" "CALL" "SECT" "CLASS" "OPER'"]
	$lstbox insert end $line
	foreach rec $byband {
		set fields [split $rec ","]
		set line [format " %3s | %3s |%10s |%5s |%4s |%10s" \
				[lindex $fields 0] \
				[lindex $fields 1] \
				[lindex $fields 2] \
				[lindex $fields 3] \
				[lindex $fields 4] \
				[lindex $fields 5]]
		$lstbox insert end $line
	}
    
	button $f.ok -text OK -command {set ok 1}
	pack $f.lst $f.ok

	wm deiconify $f
	focus $f
	grab $f
	tkwait variable ok
	grab release $f

	destroy $f

}

#=============================================================

proc ServerConfigDialog {} {
		global staCallsign staMult staClass staSection
		global socket_address socket_port
		global autosave datafile
		global config
		global dialog
	set config(Station) $staCallsign
	set config(Mult) $staMult
	set config(Class) $staClass
	set config(Section) $staSection
	set config(Autosave) $autosave
	set config(DBfile) $datafile
	set config(TCPIP_address) $socket_address
	set config(TCPIP_port) $socket_port

	set f .dialog
	if [Dialog_Create $f "Field Day Server config" -borderwidth 5] {
		frame $f.entries
		foreach field {Station Mult Class Section Autosave DBfile TCPIP_address TCPIP_port} {
			label $f.entries.l$field -text $field -anchor w
			entry $f.entries.e$field -textvariable config($field) \
				-relief sunken -bg white
			grid $f.entries.l$field $f.entries.e$field -sticky news
		}
		set b [frame $f.buttons]
		pack $f.entries $f.buttons -side top -fill x
		button $b.ok -text OK -command {set config(ok) 1}
		button $b.cancel -text Cancel -command {set config(ok) 0}
		pack $b.cancel -side left
		pack $b.ok -side right
	}

	set config(ok) 0
	Dialog_Wait $f config(ok)
	Dialog_Dismiss $f
	if {$config(ok)} {
		set staCallsign $config(Station)
		set staMult $config(Mult)
		set staClass $config(Class)
		set staSection $config(Section)
		set autosave $config(Autosave)
		set datafile $config(DBfile)
		set socket_address $config(TCPIP_address)
		set socket_port $config(TCPIP_port)
		FD_write_config
	}
}

#=============================================================

proc FDServerWindow {} {
global log contacts operator oper modes bands lblContact lblOperator
	wm withdraw .
	wm title . "Field Day Server"
	wm protocol . WM_DELETE_WINDOW FD_exit

	menu .menubar
	. config -menu .menubar
	foreach m {File View Help} {
		set $m [menu .menubar.m$m -tearoff 0]
		.menubar add cascade -label $m -menu .menubar.m$m
	}
	$File add command -label Exit -command FD_exit
	$View add command -label All -command [list FD_View All]
	foreach m $bands {
		$View add command -label "$m" -command [list FD_View $m]
	}
	$Help add command -label About -command FD_About

	frame .f1 -borderwidth 4
		label .f1.l00 -text "Sta Call:" -anchor e
		label .f1.l01 -textvariable staCallsign
		label .f1.l10 -text "    Class:" -anchor e
		label .f1.l11 -textvariable staClass
		label .f1.l20 -text "    Mult:" -anchor e
		label .f1.l21 -textvariable staMult
		pack .f1.l00 -side left -fill x -anchor e -expand true
		pack .f1.l01 -side left -anchor w
		pack .f1.l10 -side left -anchor e
		pack .f1.l11 -side left -anchor w
		pack .f1.l20 -side left -anchor e
		pack .f1.l21 -side left -anchor w -fill x -expand true
	pack .f1 -side top

	frame .f2
		label .f2.l1 -text "Cumulative Score:"
		label .f2.l2 -textvariable cumScore
		pack .f2.l1 .f2.l2 -side left -fill x -expand true
	pack .f2 -side top

	frame .f3 -borderwidth 4
	set r 0
	foreach mode $modes {
		incr r
		label .f3.m$mode -text $mode -anchor e -relief flat
		grid .f3.m$mode -sticky news -row $r -column 0
		incr r
		label .f3.op$mode -text "Oper'" -anchor e -relief flat
		grid .f3.op$mode -sticky news -row $r -column 0
	}
	set c 0
	foreach band $bands {
		incr c
		label .f3.b$band -text $band -relief flat
		grid .f3.b$band -sticky news -row 0 -column $c
	}
	set r 0
	foreach mode $modes {
		incr r
		set c 0
		foreach band $bands {
			incr c
			set lblContact($band,$mode) [label .f3.$band$mode \
						-textvariable contacts($band,$mode) \
						-width 8 -height 1 -bg white \
						-relief sunken \
						-anchor e]
			grid .f3.$band$mode -sticky news -row $r -column $c
		}
	incr r
		set c 0
		foreach band $bands {
			incr c
			set lblOperator($band,$mode) [label .f3.op$band$mode \
						-textvariable oper($band,$mode) \
						-width 8 -height 1 -bg white \
						-relief sunken \
						-anchor e]
			grid .f3.op$band$mode -sticky news -row $r -column $c
		}
	}
	pack .f3 -side top

	set log [Scrolled_Text .f4 -width 40 -height 10 -bg white]
	$log config -state disabled
	pack .f4 -side top -fill both -expand true

}

#=============================================================
proc main {argc argv} {
	global datafile autosave socket_port

	FDServerWindow

	FD_read_config
	ServerConfigDialog
	FD_init

	ServerSocket $socket_port
	SSactivate FD_server

	FD_readlog
	FD_readfile $datafile

	after $autosave FD_autosave

	wm deiconify .
}

main $argc $argv
